//////////////////////////////////////////////////////////////////
// Fir filter example.
// 
// Copyright(c) 2003-2009 Impulse Accelerated Technologies, Inc.
// 
// fir_hw.c: includes the hardware process and configuration
// function.
//
// See additional comments in fir.h.
//

#include <stdio.h>
#include "co.h"
#include "cosim_log.h"
#include "fir.h"
//#include "co_types.h"

extern void test_producer(co_stream waveform_in);
extern void test_consumer(co_stream waveform_out);

//////////////////////////////////////////////////
// This is the fir filter hardware process.
// The process accepts a stream representing a
// waveform of 32-bit samples. 
//
// This version caches n samples to create an
// n-tap filter. The values are stored in a
// circular buffer of size n.
//
// Adjust the StageDelay pragma to balance max
// frequency with pipeline rate
//
// This sample has been modified to use the Pico M501's
// 128-bit bus. 

void fir(co_stream filter_in, co_stream filter_out)
{
	picobusDataType coef[TAPS];
	IF_SIM(int samplesread = 0;)
	IF_SIM(int sampleswritten = 0;)

	picobusDataType firbuffer[TAPS];
	picobusDataType nSample;
	picobusDataType nFiltered;
	picobusDataType accum;
	int tap;

	IF_SIM( cosim_logwindow log; )
	IF_SIM ( log = cosim_logwindow_create("fir"); )

    do {	// Hardware processes run forever

		co_stream_open(filter_in, O_RDONLY,  UINT_TYPE((picobusWidth)));
		co_stream_open(filter_out, O_WRONLY, UINT_TYPE((picobusWidth)));

		// First fill the coef array with the coefficients...
		for (tap = 0; tap < TAPS; tap++) {
#pragma CO UNROLL
			co_stream_read(filter_in, &nSample, sizeof(picobusDataType));
			coef[tap] = nSample;
		
		}

		// Now fill the firbuffer array with the first n values...
		for (tap = 1; tap < TAPS; tap++) {
#pragma CO UNROLL
			co_stream_read(filter_in, &nSample, sizeof(picobusDataType));
			IF_SIM(samplesread++;)
			firbuffer[tap-1] = nSample;
		}

		// Now we have an almost full buffer and can start processing
		// the streaming waveform samples.
		//

		// Read values from the stream
		while ( co_stream_read(filter_in, &nSample, sizeof(picobusDataType)) == co_err_none ) {
#pragma CO PIPELINE
#pragma CO SET StageDelay 10
			IF_SIM(samplesread++;)
			firbuffer[TAPS-1] = nSample;

			accum = 0;
			for (tap = 0; tap < TAPS; tap++) {
#pragma CO UNROLL
				accum += firbuffer[tap] * coef[tap];
			}
			nFiltered = accum >> 2;
			co_stream_write(filter_out, &nFiltered, sizeof(picobusDataType));
			IF_SIM(sampleswritten++;)

			for (tap = 1; tap < TAPS; tap++) {
#pragma CO UNROLL
				firbuffer[tap-1] = firbuffer[tap];
			}
		}
		IF_SIM(cosim_logwindow_fwrite(log,
			"Closing fir filter process, samples read: %d, samples written: %d\n",
			samplesread, sampleswritten);)

		co_stream_close(filter_in);
		co_stream_close(filter_out);

		IF_SIM(break;)	// Only run once for desktop simulation

	} while(1);
}

//
// Impulse C configuration function
//
void config_fir(void *arg)
{
	co_stream waveform_in;
	co_stream waveform_out;
	co_process fir_process;
	co_process producer_process;
	co_process consumer_process;
	IF_SIM(cosim_logwindow_init();)

	waveform_in = co_stream_create("waveform_in", UINT_TYPE((picobusWidth)), BUFSIZE);
	waveform_out = co_stream_create("waveform_out", UINT_TYPE((picobusWidth)), BUFSIZE);

	producer_process = co_process_create("producer_process", (co_function)test_producer,
                                         1, waveform_in);
	fir_process = co_process_create("filter_process", (co_function)fir,
                                    2, waveform_in, waveform_out);
	consumer_process = co_process_create("consumer_process",(co_function)test_consumer,
                                         1, waveform_out);
	// Assign fir process to hardware elements
	co_process_config(fir_process, co_loc, "PE0");  
}

co_architecture co_initialize(int param)
{
	return(co_architecture_create("fir_arch","Generic_VHDL",config_fir,(void *)param));
}

